-- SETTINGS
local UNCURL_LOCKTIME = 8 -- Number of frames you must stay curled before you can uncurl
local BOSS_MOMENTUM_PERCENT = 75 -- Reduction factor on boss bouncing
local SHIELD_WINDOW = 7 -- Number of frames you have to let go of mid-air spin to use a shield power

-- Handle pre-frame setup
addHook("PreThinkFrame", function()
	for player in players.iterate()
		if not A_CheckExistence(player)
			continue
		end
		
		local mo = player.mo
		
		-- Init
		if player.init == nil
			A_PlayerInit(player)
		end

		if not (mo.init)
			mo.init = true
			A_ResetAbilities(player)
		end
		
		-- Dodgy flag, don't use except for bubble shield
		if not ((player.powers[pw_shield] & SH_NOSTACK) == SH_BUBBLEWRAP or (player.powers[pw_shield] & SH_NOSTACK) == SH_FLAMEAURA)
			player.pflags = $ & ~PF_SHIELDABILITY
		end
		
		-- Spindash buffs
		if player.charability2 == CA2_SPINDASH
			-- Buff minimum speed
			player.mindash = max($, 20 * FRACUNIT)
		end
		
		-- Decide whether it's time to use a shield ability
		player.shieldready = false
		if not player.usedshield
		and not (player.cmd.buttons & BT_USE)
		and player.prevbuttons & BT_USE
		and player.shieldabilitytimer > 0
		and player.shieldabilitytimer <= SHIELD_WINDOW
			player.shieldready = true
			player.cmd.buttons = $ | BT_USE
			player.pflags = $ & ~PF_USEDOWN & ~PF_THOKKED
		end
		
		-- Reset air abilities on landing, springing
		if P_IsObjectOnGround(mo)
		or (mo.eflags & MFE_SPRUNG)
		--or A_CheckNoControl(player)
			A_ResetAbilities(player)
		end
		
		-- Enter jump state mid-air to enable air abilities
		if not (P_IsObjectOnGround(mo)) and mo.state != S_PLAY_JUMP -- You're in the air but not because you jumped
		and (player.cmd.buttons & BT_JUMP) and not (player.prevbuttons & BT_JUMP) -- You've just pressed jump
		and not A_CheckNoControl(player) -- You have control
		and not (player.powers[pw_carry]) -- You're not being carried by something
		and ((mo.skin == "sonic" and not mo.doublejumped) -- If you're Sonic, you haven't used your double jump
			or (mo.skin == "tails" and mo.flycount > 0)) -- If you're Tails, you haven't burned all your fly boosts
		and not (player.pflags & PF_THOKKED) -- If anyone else, you haven't used your ability already
		and not (mo.skin == "tails" and mo.flytimer > 0 and mo.flytimer < mo.flylock) -- If you're Tails, you're not in the wait between boosts
			player.mo.state = S_PLAY_JUMP
			player.pflags = $ | PF_JUMPED
			if player.charability2 == CA2_SPINDASH
				player.pflags = $ & ~PF_NOJUMPDAMAGE
			end
		end
		
		-- Untoggled uncurling
		if not player.kr_uncurltoggle -- Uncurling is on release, not toggle
		and P_IsObjectOnGround(mo) -- You're on the ground
		and mo.state == S_PLAY_ROLL -- You're curled
		and not (A_CheckNoControl(player)) -- Control is available
		and not (mo.uncurl_lock) -- Uncurling is not locked
		and not (player.cmd.buttons & BT_USE) -- You're not holding spin
			A_Uncurl(player)
		end
	end
end)

-- Toggled uncurling and aircurling for all characters
addHook("SpinSpecial", function(player)
	for player in players.iterate()
		-- Have to exist to uncurl
		if not A_CheckExistence(player)
			return false
		end
		
		local mo = player.mo
		
		-- On ground, attempt toggled uncurl
		if P_IsObjectOnGround(mo)
			if mo.state == S_PLAY_ROLL -- Must be curled
			and player.kr_uncurltoggle -- Uncurl toggle must be on
			and not A_CheckNoControl(player) -- Control must be available
			and not mo.uncurl_lock -- Uncurling can't be locked
			and not (player.prevbuttons & BT_USE) -- Can't be already holding spin
				A_Uncurl(player)
				return true
			end
		else
			if not A_CheckNoControl(player)
			and (player.pflags & PF_JUMPED)
				player.shieldabilitytimer = $+1
			end
			
			-- Try to curl in the air
			if player.charability2 == CA2_SPINDASH -- Must be able to spin
			and mo.state != S_PLAY_ROLL -- Can't already be curled up
			and mo.state != S_PLAY_JUMP -- Can't already be curled up
			and not (player.prevbuttons & BT_USE) -- Must be pressing spin, not holding it
			and not (player.powers[pw_carry]) -- Can't be being carried by something
				A_Aircurl(player)
				return true
			end
		end
	end
end)

addHook("ShieldSpecial", function(player)
	-- Use shield power if we decided it's ready earlier
	if player.shieldready
		-- Bubble shield can be chained
		if (player.powers[pw_shield] & SH_NOSTACK) == SH_BUBBLEWRAP or (player.powers[pw_shield] & SH_NOSTACK) == SH_FLAMEAURA
			player.pflags = $ | PF_SHIELDABILITY
		end
		
		if (player.powers[pw_shield] & SH_NOSTACK) != SH_BUBBLEWRAP
			player.usedshield = true
		end
		return false
	-- Otherwise don't
	else
		return true
	end
end)

-- Various controls for all characters
addHook("PlayerThink", function(player)
	-- Big list of reasons to bail out early
	if not A_CheckExistence(player)
	or not A_CheckStockChar(player)
	or not (player.playerstate == PST_LIVE)
	or player.exiting
	or player.powers[pw_nocontrol]
	or P_PlayerInPain(player)
		return
	end
	
	local mo = player.mo
	
	-- General moveset tweaks
	
	-- Uncurl lock timer
	-- If you're rolling, count down the timer
	if P_IsObjectOnGround(mo) and mo.state == S_PLAY_ROLL 
		if mo.uncurl_lock
			mo.uncurl_lock = max(0, $ - 1)
		end
	-- If you're not, reset the timer
	else
		mo.uncurl_lock = UNCURL_LOCKTIME
	end
	
	-- Full air control when jumping out of a roll for all characters
	if player.pflags & PF_SPINNING and not P_IsObjectOnGround(mo)
		if player.pflags & PF_JUMPED
			player.pflags = $ & ~PF_SPINNING
		end
	end
end)

-- Regain abilities after hitting monitor/enemy/boss
addHook("MobjMoveCollide", function(mo, touched)
	local player = mo.player
	
	if not A_CheckExistence(player)
	or not (touched.flags & (MF_MONITOR|MF_ENEMY|MF_BOSS))
	or not (touched.flags & MF_SHOOTABLE)
	or (touched.flags & MF_MISSILE)
	or not A_CheckCollisionHeight(mo, touched)
		return
	end
	
	A_ResetAbilities(player, true)
		
	-- Reduce x and y velocity when bouncing off a boss/facestabber/etc
	if (touched.flags & MF_BOSS) or (touched.health > 1)
		mo.momx = $ * BOSS_MOMENTUM_PERCENT / 100
		mo.momy = $ * BOSS_MOMENTUM_PERCENT / 100
	end
end, MT_PLAYER)

addHook("PostThinkFrame", function()
	for player in players.iterate()
		-- Existence check
		if not A_CheckExistence(player)
			continue
		end
		
		local mo = player.mo
		
		-- Cut the aircurl sound early if you're not actually curled
		if mo.aircurled and not (mo.state == S_PLAY_JUMP or mo.state == S_PLAY_ROLL)
			S_StopSoundByID(mo, sfx_curl)
		end
		
		player.prevbuttons = player.cmd.buttons
		if player.powers[pw_carry]
			player.prevbuttons = $ | BT_JUMP
		end
		
		-- Reset the aircurled check
		mo.aircurled = false
		
		-- Reset the shield ability timing counter
		if not (player.cmd.buttons & BT_USE)
		or P_IsObjectOnGround(mo)
		or A_CheckNoControl(player)
		or player.shieldready
			player.shieldabilitytimer = 0
		end
	end
end)
